//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
//              ____  ____   __  ____  _  _  __  ____             //
//             (  _ \(  _ \ /  \(  _ \/ )( \(  )/ ___)            //
//              ) __/ )   /(  O )) _ (\ \/ / )( \___ \            //
//             (__)  (__\_) \__/(____/ \__/ (__)(____/            //
//                                                                //
//                  Written by Kristi Potter 2011                 //
//                 SCI Institute, Unversity of Utah               //
//                       kpotter@sci.utah.edu                     //
// Tool for comparing and exploring Probability Density Functions //
//                                                                //
// * * * * * * * * * * * * LICENSE (BSD) * * * * * * * * * * * * *//
// Copyright (c) [2011] [Kristin Potter]. All Rights Reserved.    //
// Redistribution and use in source and binary forms, with or     //
// without modification, are permitted provided that the          // 
// following conditions are met:                                  //
//                                                                //
// 1. Redistributions of source code must retain the above        //
// copyright notice, this list of conditions and the following    //
// disclaimer.                                                    //
//                                                                //
// 2. Redistributions in binary form must reproduce the above     //
// copyright notice, this list of conditions and the following    //
// disclaimer in the documentation and/or other materials         //
// provided with the distribution.                                //
//                                                                //
// 3. The name of the author may not be used to endorse or        //
// promote products derived from this software without specific   //
// prior written permission.                                      //
//                                                                //
// THIS SOFTWARE IS PROVIDED BY [KPOTTER] "AS IS" AND ANY         //
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,  //
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A    //
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE       //
// AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,//
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED//
// TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  //
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND //
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT        //
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING //
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF //
// THE POSSIBILITY OF SUCH DAMAGE.                                //
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
import processing.opengl.*;
import javax.media.opengl.GL;
import controlP5.*;
import picking.*;

// Global variables
GL gl;
Picker picker;
ControlP5 controlP5;

DataReader reader;
DataObject object;
DistObject distrib;
PdfDistance pdfDiff;

OverlayScene overlay;
SideByScene  sideBy;

DensityFuncDisplay dataDisplay;
DensityFuncDisplay distDisplay;

Point currentPt;
Point _locate= new Point();
int numBins = 256;

KPButton loadKP, updateKP, resetKP, meshKP, overlayKP, diffKP, distKP;

Numberbox binBox;
Toggle toggle;
Textlabel cdfLabel, pdfLabel;
Textlabel diffLabel, distLabel;
Textlabel binLabel;

ControlGroup binGroup, binGroup2, diffGroup;
ControlGroup optionsBox, radioGroup, paramsGroup;
ControlGroup distGroup, diffBox;
Button okButton, cancelButton;
DropdownList distOptionList;
RadioButton distOptionRadio;

Point trans, rot;

boolean rotating, zooming, selecting, moving;
boolean dataLoaded = false;
boolean oneDColor = false;
boolean noTransform = false;
boolean drawMesh = true;
boolean simultaneous = true;

// * * * Set up the environment (called once) * * * //
void setup() {

    // WEIRD NOTE: Everything above the size() call will be called twice!
    size(1000, 500, OPENGL);
    frameRate(25);
    gl=((PGraphicsOpenGL)g).gl;
    picker = new Picker(this);

    // Initialize the mouse-based movements
    trans = new Point(); rot = new Point();
    rotating=zooming=selecting=moving=false;
    
    // Initialize the objects
    object  = new DataObject();
    distrib = new DistObject(numBins);
    overlay = new OverlayScene();
    sideBy  = new SideByScene();
    dataDisplay = new DensityFuncDisplay("LEFT");
    distDisplay = new DensityFuncDisplay();
  
    // Setup the GUI elements
    setupController();
}

// * * * Draw the scene (called everytime) * * * //
void draw() {
    hint(ENABLE_DEPTH_TEST);
    background(192);
  
    // Set the specular color of lights that follow
    lightSpecular(255, 255, 255);
    directionalLight(204, 204, 204, 0, 0, -1);
    ambientLight(102, 102, 102);
    shininess(1.0);
    
    // Draw the meshes
    if(dataLoaded){
	// Draw the mesh & colorbar
	if(simultaneous)
	    overlay.drawScene(picker, object, drawMesh, trans, rot);
	else
	    sideBy.drawScene(picker, object, drawMesh, trans, rot);
	
	// Draw the density functions
	dataDisplay.drawDensityFunc(object, currentPt.idx(), distrib);
	distDisplay.drawDensityFunc(distrib, currentPt.idx(), object);
    }

    hint(DISABLE_DEPTH_TEST);
}

// * * * Set up the GUI controller * * * //
void setupController(){
    
    // Create a new controller and set fonts & colors
    controlP5 = new ControlP5(this);
    KPCustomize(controlP5);
   
    // Add the GUI elements
    loadKP = new KPButton(controlP5, "loadGroup", "Load Data", 4, 4, 75, 25, false);

    // Add the bin box
    binGroup = controlP5.addGroup("binGroup", 84, 2, 143);
    binGroup.setBackgroundHeight(30);       
    binGroup.setBackgroundColor(color(135));
    binGroup.hideBar(); 
    binGroup2 = controlP5.addGroup("binGroup2", 86, 4, 91);
    binGroup2.setBackgroundHeight(26);       
    binGroup2.setBackgroundColor(color(135));
    binGroup2.hideBar(); 
    binLabel =  controlP5.addTextlabel("binLabel","Number of Bins:", 88, 12);
    binLabel.setColorValue(color(15));
    binBox = controlP5.addNumberbox("Bins",object.getNumBins(),180,4,45,25);
    binBox.setDirection(Controller.HORIZONTAL); binBox.setLabel("");
    binBox.setMin(0);

    // Update number of bins button
    updateKP = new KPButton(controlP5, "updateGroup", "Update", 227, 4, 57, 25);
  
    // Reset viewing
    resetKP = new KPButton(controlP5, "resetGroup", "Reset Camera",  width/2-138, 4, 102, 25);
   
    // Toggle mesh display
    meshKP  = new KPButton(controlP5, "meshGroup", "Mesh on/off",  width/2-29, 4, 88, 25);

    // Toggle overlay
    overlayKP = new KPButton(controlP5, "overlayGroup","Overlay on/off", width/2+66, 4, 112, 25);

    // Change the measure
    diffKP = new KPButton(controlP5, "diffGroup2","Change",  width-144, 4, 60, 25);
    diffGroup = controlP5.addGroup("diffGroup", width-220, 4, 74);    
    diffGroup.setBackgroundHeight(26);       
    diffGroup.setBackgroundColor(color(135));
    diffGroup.hideBar(); 
    diffLabel = controlP5.addTextlabel("diffLabel",   "L1 Norm", width-205, 12);
    diffLabel.setColorValue(color(15)); 
    diffGroup.hide(); 
    diffLabel.hide();
    
    // Toggle cdf/pdf
    int x = width-200; int y = height-100;
    toggle = controlP5.addToggle("Density Function",true,x,y,60,15);
    toggle.setColorBackground(color(140));
    toggle.setColorActive(color(178,13,13));   
    toggle.setMode(ControlP5.SWITCH);
    toggle.setLabelVisible(false);
    pdfLabel = controlP5.addTextlabel("pdfLabel","PDF",  x+3, y-10);
    cdfLabel = controlP5.addTextlabel("cdfLabel","CDF", x+33, y-10);
    pdfLabel.setColorValue(color(15)); cdfLabel.setColorValue(color(15));
    toggle.hide(); pdfLabel.hide(); cdfLabel.hide();

    //Distribution options box
    distKP = new KPButton(controlP5, "distGroup2", "Change/Modify", 215, y-5, 108, 25);  
    distGroup  = controlP5.addGroup("distGroup",140, y-5, 75);
    distGroup.setBackgroundHeight(26);       
    distGroup.setBackgroundColor(color(135));
    distGroup.hideBar();                     
    distLabel = controlP5.addTextlabel("distLabel", distrib.getLabel(), 155, y+3);
    distLabel.setColorValue(color(15)); 
    distLabel.hide();    
    distGroup.hide();
   
    optionsBox = controlP5.addGroup("messageBox",width/2 - 90,185,195);
    optionsBox.setBackgroundHeight(106);
    optionsBox.setBackgroundColor(color(0,100));
    optionsBox.hideBar(); optionsBox.hide();

    radioGroup = controlP5.addGroup("", 50, 8, 100);
    radioGroup.moveTo(optionsBox);
    radioGroup.setBackgroundHeight(60);
    radioGroup.setBackgroundColor(color(145));
    radioGroup.hideBar();

    distOptionRadio = controlP5.addRadioButton("radioButton",55,19);
    distOptionRadio.setColorActive(color(178, 13, 13));
    distOptionRadio.setSpacingColumn(50);
    distOptionRadio.addItem("Uniform", DistObject.UNIFORM); 
    distOptionRadio.addItem("Gaussian", DistObject.GAUSSIAN);
    distOptionRadio.addItem("Beta", DistObject.BETA);
    distOptionRadio.moveTo(optionsBox);  
   
    okButton = controlP5.addButton("buttonOK",0,36,73,57,25);
    okButton.moveTo(optionsBox);
    okButton.captionLabel().style().setMarginLeft(16);
    okButton.setCaptionLabel("OK");
    
    cancelButton = controlP5.addButton("buttonCancel",0,106,73,57,25);
    cancelButton.moveTo(optionsBox);
    cancelButton.setCaptionLabel("Cancel");

    distOptionRadio.activate(distrib.getType());
}

// * * * Handle gui events * * * //
public void controlEvent(ControlEvent theEvent) {
    
    boolean uniOn = false;
    boolean gaussOn = false;
    
    // Check if we have a controller
    if(theEvent.isController()){

	// Load the data set
	if(theEvent.controller().name() == "Load Data") {
	    String folderPath = selectFolder();
	    if (folderPath == null) {
		// If a folder was not selected
		println("ERROR! No folder was selected...");
	    }
	    else {
		
		// Create a new data reader & read in data
		reader = new DataReader(folderPath);
		object.setGeometry(reader.readGeometry());
		object.setConnectivity(reader.readConnectivity());
		reader.generateHistogram(numBins, object);
		distrib.generateDistribution(object);
		pdfDiff = new PdfDistance(object, distrib);
		
		setColorbarPt();
		updateColormap();

		dataLoaded = true;
		toggle.show();
		pdfLabel.show(); cdfLabel.show();
		distKP.show(); distGroup.show();  setDistLabel();
		diffKP.show(); diffGroup.show(); diffLabel.show();
		resetKP.show();
		meshKP.show();
		overlayKP.show();
	    }
	}
	// Update the data
	if(theEvent.controller().name() == "Update") {
	    if(reader != null) reader.generateHistogram(numBins, object);    
	    distrib.generateDistribution(object);
	    pdfDiff = new PdfDistance(object, distrib);
	    setColorbarPt();
	    updateColormap();
	    updateKP.hide();
	}
	// CDF vs PDF
	if(theEvent.controller().name() == "Density Function") {
	    dataDisplay.toggleDisplay();
	    distDisplay.toggleDisplay();
	}
	// Show options box
	if(theEvent.controller().name() == "Change/Modify") {
	    optionsBox.show();
	    noTransform = true;
	}
	// Cancel options box changes
	if(theEvent.controller() == cancelButton){
	    optionsBox.hide();
	    noTransform = false;
	}
	// Ok button
	if(theEvent.controller() == okButton){
	    distrib.changeType((int)distOptionRadio.value(), object);
	    pdfDiff = new PdfDistance(object, distrib, pdfDiff.getType());
	    updateColormap();
	    setColorbarPt();
	    optionsBox.hide();
	    setDistLabel();
	    noTransform = false;
	}
	// Difference button
	if(theEvent.controller().name() == "Change"){
	    pdfDiff.toggleMeasure(); 
	    updateColormap(); 
	    setColorbarPt();
	    diffLabel.setValue(pdfDiff.getTypeName());
	}
	// Reset button
	if(theEvent.controller().name() == "Reset Camera"){
	    trans.setZero(); rot.setZero();
	}
	// Mesh button
	if(theEvent.controller().name() == "Mesh on/off"){
	    drawMesh = !drawMesh;
	}
	// Overlay button
	if(theEvent.controller().name() == "Overlay on/off"){
	    simultaneous = !simultaneous;
	}
    }
    else if(theEvent.isGroup()){
	// Distribution radio button
	if(theEvent.group() == distOptionRadio){
	}    
    }
}

// * * * Monitor the number of bins * * * //
void Bins(int bins) {
    numBins = bins; 
    if (numBins != object.getNumBins() && dataLoaded)  {
	updateKP.show();
	//hideUpdate = false;
	//	updateGroup.show(); updateButton.show();
    }
    else{
	updateKP.hide();
    	//hideUpdate = true;
	//	updateGroup.hide(); updateButton.hide();
    }
}

// * * * Change the view based on the mouse  * * * //
void mouseDragged() {
    
    if(!noTransform){
	// If we are selecting, translate the pointer, not the scene
	if(selecting) {
	    
	    // Translate the pointer
	    if(mouseButton == LEFT && ! zooming){
		_locate.x(_locate.x()-(pmouseX-mouseX)); _locate.y(_locate.y()-(pmouseY-mouseY)); }
	    // Zoom the camera
	    if(mouseButton == CENTER || zooming){ 	_locate.z(_locate.z()+(pmouseX-mouseX)); } 
	    
	    setColorbarPt();
	}
	
	// Else, transform the scene
	else {
	    if(!moving){
		float diffX = pmouseX - mouseX;  float diffY = pmouseY - mouseY;
		// Translate the camera
		if(mouseButton == LEFT && !rotating && !zooming){
		    trans.x(trans.x() - diffX);	 trans.y(trans.y() - diffY);
		}
		// Rotate the camera
		else if(mouseButton == RIGHT || rotating){
		    rot.x(rot.x() + diffY);      rot.y(rot.y() - diffX);
		}
		// Zoom the camera
		if(mouseButton == CENTER || zooming){
		    trans.z(trans.z() + diffX);  trans.z(trans.z() - diffY);
		} 
	    }
	}
    }
}

// * * * The mouse is pressed * * * //
void mousePressed() {
    int id = picker.get(mouseX, mouseY);
    if (id > -1) {
	selecting = true;
	overlay.setSelected(selecting); 
	sideBy.setSelected(selecting);
    } 
   if(mouseX > 180 & mouseX < 226 & mouseY > 5 & mouseY < 31)
	moving = true;
}

// * * * The mouse is released * * * //
void mouseReleased(){
    selecting = false;
    overlay.setSelected(selecting);
    sideBy.setSelected(selecting);
    moving = false;
}

// * * * Register key press * * * //
void keyPressed() {
    
    if(key == CODED){ }
    else if(key == 'x') rotating = true;
    else if(key == 'z') zooming = true;
    else if(key == ' ') { pdfDiff.toggleMeasure(); updateColormap(); }
    else if(key == 'h') { trans.setZero(); rot.setZero();}
}

// * * * Release the keys * * * //
void keyReleased(){ rotating = false; zooming = false; }

// * * * Update the colormap * * * //
void updateColormap() {
    
    // If we have a uniform distribution
    if(distrib.getType() == DistObject.UNIFORM){
	overlay.setColorData(pdfDiff.getDistances(),   pdfDiff.getIntervals(), 
			     pdfDiff.getMinDist(),     pdfDiff.getMaxDist(),
			     pdfDiff.getMinInterval(), pdfDiff.getMaxInterval(),
			     pdfDiff.getTypeName(),    "Interval\n Width");

	sideBy.setColorData(pdfDiff.getDistances(),    pdfDiff.getIntervals(), 
			     pdfDiff.getMinDist(),     pdfDiff.getMaxDist(),
			     pdfDiff.getMinInterval(), pdfDiff.getMaxInterval(),
			     pdfDiff.getTypeName(),    "Interval Width");
    }
    else{
	overlay.setColorData(pdfDiff.getDistances(), object.getStdDevs(), 
			     pdfDiff.getMinDist(),   pdfDiff.getMaxDist(), 
			     pdfDiff.getMinStd(),    pdfDiff.getMaxStd(),
			     pdfDiff.getTypeName(),  "Variance");

	sideBy.setColorData(pdfDiff.getDistances(),  object.getStdDevs(), 
			     pdfDiff.getMinDist(),   pdfDiff.getMaxDist(), 
			     pdfDiff.getMinStd(),    pdfDiff.getMaxStd(),
			     pdfDiff.getTypeName(),  "Variance");
    }
}

// * * * Set the current point on the colorbar * * * //
void setColorbarPt(){
    currentPt = object.getClosestPt(_locate);
    Point twoPt = new Point();
    if(distrib.getType() == DistObject.UNIFORM)
	twoPt = new Point((float)pdfDiff.getDistance(currentPt.idx()), 
			  (float)pdfDiff.getInterval(currentPt.idx()));
    else 
	twoPt = new Point((float)pdfDiff.getDistance(currentPt.idx()),
			  (float)object.getStdDev(currentPt.idx()));
    overlay.setCurrentPt(currentPt, twoPt);
    sideBy.setCurrentPt( currentPt, twoPt);     
}

// * * * Set the distribution label * * * //
void setDistLabel(){
    distLabel.setValue(distrib.getLabel());
    distLabel.show();
}
